from __future__ import annotations
from array import array
import hashlib
import asyncio
from sqlite3.dbapi2 import Error
from common.messaging.message import VumosMessage, VumosMessageProcessed
from nats.aio.client import Client as NATS
import sched
import json
import time
from typing import Callable, Coroutine, List, TypedDict, Any
import sqlite3
import uuid
import re
from datetime import datetime
import os
from enum import Enum
import logging


broadcast_subject = "broadcast"


class VumosMessageType(Enum):
    REQUEST = "request"
    RESPONSE = "response"


class VumosServiceStatus:
    code: str
    message: str

    def __init__(self, code: str, message: str) -> None:
        self.code = code.lower()
        self.message = message


class VumosParameterValue(TypedDict):
    type: str
    default: Any


class VumosParameter(TypedDict):
    name: str
    description: str
    key: str
    value: VumosParameterValue


class VumosAction(TypedDict):
    name: str
    description: str
    key: str
    arguments: List[VumosParameter]


class VumosService:
    def __init__(self,
                 name: str, description: str,
                 parameters: List[VumosParameter] = [],
                 actions: List[VumosAction] = [],
                 nats_callback: Callable[[
                     VumosService, VumosMessage], None] = None,
                 status_expiry: float = 60,
                 database_file='persistent/vumos-service.db') -> None:
        self.data_mtype_regex = re.compile(r'data_([a-z]+)')

        self.id = os.environ.get('VUMOS_ID') or str(uuid.uuid4())
        self.scheduler = sched.scheduler(time.time, time.sleep)
        self.running = True

        self.actions = actions

        self.handlers = []
        self.action_handlers = {}

        # do nothing on restart for now, used only for periodic
        self.action_handlers['restart'] = self.restart

        ######################
        ## Sets up database ##
        ######################
        self.database = sqlite3.connect(database_file)

        # Creates configuration table and load stored values
        self.database.execute('''
        CREATE TABLE IF NOT EXISTS configuration (
            key TEXT PRIMARY KEY,
            value TEXT
        )''')

        valid_keys = list(map(lambda p: p['key'], parameters))

        for key, value in self.database.execute('SELECT key, value FROM configuration'):
            if not key in valid_keys:
                self.database.execute(
                    'DELETE FROM configuration WHERE key = ?', (key,))

        self.database.commit()

        for config in parameters:
            key = config["key"]
            default = config["value"]["default"]
            try:
                self.get_config(key)
            except Error:
                self.set_config(key, default)

        #############################
        ## Sets up node parameters ##
        #############################
        self.name = name
        self.description = description
        self.parameters = parameters
        self.status_expiry = status_expiry

        self.nats_callback = nats_callback

        self.set_status(VumosServiceStatus(
            "red", f"[ERROR] Service still has no status set"))

    def set_status(self, status: VumosServiceStatus):
        '''
        This function sets the current status for the service to send to managers.

        Parameters:
            status (VumosServiceStatus): The current status of this service
        '''
        logging.debug(f"Status set to: {status.message}")
        self.status = status

    def set_config(self, key: str, value: Any) -> Any:
        '''
        This function sets the current value of the configuration of a given key for this service.

        Parameters:
            key   (str): The key of the configuration
            value (any): The value of the configuration

        Returns:
            Any: The value of the configuration
        '''
        self.database.execute(
            'REPLACE INTO configuration VALUES (?, ?)', (key, json.dumps(value)))
        self.database.commit()

        return value

    def get_config(self, key: str) -> Any:
        '''
        This function returns the current value of the configuration of a given key for this service.

        Parameters:
            key (str): The key of the configuration

        Returns:
            Any: The value set by the manager or default configuration
        '''
        for (value,) in self.database.execute('SELECT value FROM configuration WHERE key = ?', (key,)):
            return json.loads(value)

        raise Error("Configuration does not exist")

    def on_action(self, key: str, handler: Callable[[VumosService, List[VumosParameter]], Coroutine] = None):
        '''
        This function adds an action handler to the service.

        Parameters:
            key (str): The key of the action
            handler (callable): A handler function. It will be called with a service and parameters arguments
        '''
        self.action_handlers[key] = handler

    def on(self, filter: Callable[[VumosMessage], bool], handler: Callable[[VumosService, VumosMessage], Coroutine] = None):
        '''
        This function adds a data handler to the service.

        Parameters:
            filter (callable): A filter function, checks if the message should or not be processed by the handler
            handler (callable): A handler function. It will be called with a service and message arguments
        '''
        self.handlers.append((filter, handler))

    async def connect(self, uri=os.getenv("NATS_URI", "nats://127.0.0.1:4222")) -> None:
        '''
        This method connects the service to the NATS messaging service, sets up subscriptions, and sends the HELLO message.

        Parameters:
        '''
        # Connection to Nats
        self.nats = NATS()
        await self.nats.connect(uri)

        # Listen for messages

        async def handler(msg):
            await self._message_callback(msg)

        await self.nats.subscribe(broadcast_subject, cb=handler)
        await self.nats.subscribe(self.id, cb=handler)

        # Send Hello
        await self._send_hello()
        await self._send_cchanged()

    def loop(self) -> None:
        '''
        This method starts listening for messages and sending status updates to the backbone.
        '''

        # Status update loop
        async def status_update_loop():
            while self.running:
                await self._send_status()
                await asyncio.sleep(0.75 * self.status_expiry)

        scheduled = asyncio.get_event_loop().create_task(status_update_loop())
        asyncio.get_event_loop().run_until_complete(scheduled)

    # Listen for message callbacks

    async def _message_callback(self, msg) -> None:
        '''
        This method handles an incoming message from the network.
        '''
        message: VumosMessage = json.loads(msg.data.decode('utf-8'))

        # Base message parameters
        m_id = message["id"]
        m_type = message["message"]
        m_source = message["source"]
        m_mode = message["mode"]
        m_processed = message["processed"]
        m_data = message["data"]

        # Skip message if is repeated
        hash = hashlib.md5(json.dumps(m_data).encode('utf-8')).hexdigest()
        for processed in m_processed:
            if processed["hash"] == hash and processed["module"] == self.id:
                return

        # Ignore message if from this or other service
        if m_id == self.id:
            return

        logging.debug("========== Receive ==========")
        logging.debug(json.dumps(message, indent=2))
        type_converters = {
            'string': str,
            'integer': int,
            'float': float
        }

        if m_type == "hello":
            # On hello message
            #
            # broadcast: Reply directly
            # directed: Don't do anything
            #
            # always: Send current configuration to managers
            if m_mode == "broadcast":
                await self._send_hello(msg.reply)

            if m_source == "manager":
                await self._send_status(msg.reply)
                await self._send_cchanged(msg.reply)

        elif m_type == "configuration_change":
            # On configuration change message
            #
            # broadcast: Don't do anything
            # directed: Set configuration changes
            #
            # always: Send current configuration to managers

            if m_mode == "targeted":
                for config in m_data['configurations']:
                    c_key = config['key']
                    c_value = config['value']['current']

                    try:
                        self.get_config(c_key)
                    except Error:
                        logging.warning(f'Ignoring unknown config {c_key}')
                        continue

                    try:
                        self.set_config(
                            c_key, type_converters[config['value']['type']](c_value))
                    except Exception as e:
                        logging.error(
                            f"Failed to convert value '{c_value}' [{type(c_value)}] to type {config['value']['type']}")
                        logging.error(e)

                await self._send_cchanged()
                logging.info("Received configuration update")
                await asyncio.sleep(2) # this is because canceling the task for some reason crashes NATS, so wait a bit so we don't lose the CCHANGED message
                action_handler = self.action_handlers['restart']
                await action_handler()
        elif m_type == "execute_action":
            # On execute action message
            #
            # broadcast: Don't do anything
            # directed: Execute action

            # Parse action arguments
            arguments = {}

            for argument in m_data['arguments']:
                c_key = argument['key']
                c_value = argument['value']['current']

                arguments[c_key] = type_converters[argument['value']
                                                   ['type']](c_value)

            # Run handler
            action_handler = self.action_handlers[m_data['key']]

            if action_handler:
                await action_handler(self, arguments)

        elif m_type == "status_update":
            # On status update, do nothing
            pass
        elif self.nats_callback:
            self.nats_callback(self, message)

        # Run handlers for this message
        data_match = self.data_mtype_regex.fullmatch(m_type)

        if data_match:
            for filter, handler in self.handlers:
                if filter(message):
                    await handler(self, message)

    ###########################
    # Message sending methods #
    ###########################
    async def send_message(self, message: str, data: dict, to: str = None, type: VumosMessageType = VumosMessageType.REQUEST, processed: List[VumosMessageProcessed] = [], custom_id = None) -> None:
        new_processed = list(processed)

        new_hash = hashlib.md5(json.dumps(data).encode('utf-8')).hexdigest()
        new_ts = datetime.utcnow().isoformat()

        found = False
        for module in new_processed:
            if module["module"] == self.id:
                module["hash"] = new_hash
                module["timestamp"] = new_ts
                found = True
                break

        if not found:
            new_processed.append({
                "module": self.id,
                "hash": new_hash,
                "timestamp": new_ts
            })

        message: VumosMessage = {
            "id": self.id,
            "message": message,
            "source": "service",
            "type": type.value,
            "mode": "broadcast" if to is None else "targeted",
            "processed": new_processed,
            "data": data
        }

        logging.debug("========== Sending ==========")
        logging.debug(json.dumps(message, indent=2))

        subject = to or broadcast_subject
        payload = json.dumps(message).encode('utf-8')

        await self.nats.publish(subject, payload=payload)

    async def send_target_data(self, ip_address: str, domains: array(str) = [], extra: Any = None, to: str = None) -> None:
        await self.send_message("data_target", {
            "ip_address": ip_address,
            "domains": domains,
            "extra": extra
        }, to=to)

    async def send_path_data(self, data: object = {}, to: str = None) -> None:
        await self.send_message("data_path", data)

    async def send_issue_data(self, ip_address: str, port: int, name: str, fixed: bool, description: str = None, cve: str = None, cvss: str = None, extra: Any = None, notes: str = None, to: str = None) -> None:
        data = {
            "ip_address": ip_address,
            "port": port,
            "name": name,
            "fixed": fixed,
        }

        if description:
            data['description'] = description
        if cve:
            data['cve'] = cve
        if cvss:
            data['cvss'] = cvss
        if extra:
            data['extra'] = extra
        if notes:
            data['notes'] = notes

        await self.send_message("data_issue", data, to=to)


    async def send_service_data(self, ip_address: str, port: int, name: str = None, protocol: str = None, version: str = None, extra: Any = None, to: str = None) -> None:
        data = {
            "ip_address": ip_address,
            "port": port,
        }

        if name:
            data['name'] = name
        if protocol:
            data['protocol'] = protocol
        if version:
            data['version'] = version
        if extra:
            data['extra'] = extra

        await self.send_message("data_service", data, to=to)

    async def send_target(self, to: str = None) -> None:
        await self.send_message("", {
            "name": self.name,
            "description": self.description,
            "status_expiry": self.status_expiry
        }, to=to)

    async def _send_hello(self, to: str = None) -> None:
        await self.send_message("hello", {
            "name": self.name,
            "description": self.description,
            "status_expiry": self.status_expiry
        }, to=to)

    async def _send_cchanged(self, to: str = None) -> None:
        def apply_current_value(p: VumosParameter):
            data = dict(p)
            data["value"]["current"] = self.get_config(p["key"])
            return data

        await self.send_message("configuration_changed", {
            "configurations": list(map(apply_current_value, self.parameters)),
            "actions": self.actions
        }, to=to)

    async def _send_status(self, to: str = None) -> None:
        await self.send_message("status_update", {
            "code": self.status.code,
            "message": self.status.message
        }, to=to)

    async def restart(self) -> None:
        pass