import { DateTime } from 'luxon';
import { Entity, PrimaryGeneratedColumn, Column, CreateDateColumn, UpdateDateColumn, DeleteDateColumn } from 'typeorm';

interface StringValue {
  type: 'string';
  current: string;
  default: string;
}

interface NumberValue {
  type: 'number';
  current: number;
  default: number;
}

export type ModuleConfigurationValue = StringValue | NumberValue;

export class ModuleConfiguration {
  name: string;
  description: string;
  key: string;
  value: ModuleConfigurationValue;
}

@Entity('modules')
export class Module {
  @PrimaryGeneratedColumn('uuid')
  id: string;

  @Column({
    comment: 'Unique identifier generated by the module',
    unique: true,
  })
  uid: string;

  @Column({
    comment: 'Module name',
  })
  name: string;

  @Column({
    comment: 'Module description',
  })
  description: string;

  @Column({
    comment: 'Module nickname',
    nullable: true,
  })
  nickname: string;

  @Column({
    comment: 'User notes for this module',
    nullable: true,
  })
  notes?: string;

  @Column({ comment: 'Module status', nullable: true })
  public status_code?: string;

  @Column({ comment: 'Module status message', nullable: true })
  public status_message?: string;

  @Column({ comment: 'Last module status update timestamp', nullable: true })
  public status_last_update?: Date;

  @Column({ comment: 'Time period in seconds the status stays valid' })
  public status_expiry: number;

  public get status_valid(): boolean {
    return (
      DateTime.utc().valueOf() <
      DateTime.fromJSDate(this.status_last_update).plus({ seconds: this.status_expiry }).valueOf()
    );
  }

  @Column('jsonb', {
    comment: 'Configurations for this module',
  })
  configurations: ModuleConfiguration[] = [];

  // Relationships

  // Auto timestamps
  @Column()
  @CreateDateColumn()
  created_at: Date;

  @Column()
  @UpdateDateColumn()
  updated_at: Date;

  @Column({ nullable: true })
  @DeleteDateColumn()
  deleted_at?: Date;

  // For dealing with unconventional values
  public toJSON() {
    console.log(DateTime.fromJSDate(this.status_last_update).plus({ seconds: this.status_expiry }).toISO());
    console.log(DateTime.utc().toISO());
    return {
      ...this,
      status_valid: this.status_valid,
    };
  }
}
